/*
 * Copyright (c) 1999 SuSE GmbH Nuernberg, Germany.
 *
 * Author(s): Marius Tomaschewski <mt@suse.de>, 1999
 *
 * Licence is GPL, so you may find it in /usr/src/linux/COPYING.
 * Otherwise mail me and I will sent it to you.
 *
 * $Id: squid_ldapauth.c,v 1.3 2000/01/27 17:26:43 mt Exp $
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>

#include <sys/param.h>
#include <netdb.h>

#include <syslog.h>
#include <signal.h>
#include <crypt.h>

#include <lber.h>
#include <ldap.h>

#define      ldap_crypt_prefix   "{crypt}"
const size_t crypt_prefix_size = sizeof(ldap_crypt_prefix)-1;

#ifndef CONFIG_FILENAME
#define CONFIG_FILENAME   "/etc/squid_ldapauth.conf"
#endif
#define SQUID_ERR         "ERR\n"
#define SQUID_OK          "OK\n"

struct Conf {
  char  *key;
  char  *val;
} conf[] = {
  {"ldap-server",      0},
  {"ldap-port",        0},
  {"ldap-suffix",      0},
  {"ldap-filter",      0},
  {"ldap-passwdfield", 0},
  {"ldap-binddn",      0},
  {"ldap-password",    0},
  {0}
};

LDAP   *ldap      = 0;
char   **attrs    = 0;
char   *appname   = 0;

int    log_queries= 0;
int    be_verbose = 0;
int    use_syslog = 1;

void  usage();
void  quit(int status);

void  squidReply(char *reply, char *user);

int   checkPass(char *user, char *spass, char *lpass);
int   ldapQueryPass(LDAP *ldap, char *suffix, char *filter,
                    char **attrs, char *user, char *pass);
LDAP* ldapConnect(char *server, int port, char *suffix);

void  initConfig();
void  freeConfig();
int   readConfig(char *filename);

int   getfqhostname(char *fqhost, size_t n);
int   getdnsdomain(char *domain, size_t n);


/* ---------------------------------------------------------------------- */
int main(int argc, char **argv)
{
  char   buf[BUFSIZ] = {0};
  char   *server = 0;
  char   *suffix = 0;
  char   *filter = 0;
  int    i, port = LDAP_PORT;
  char   *user, *pass;

  if( (appname = strrchr(argv[0], '/'))) appname++;
  else appname = argv[0];

  for(i=1; i<argc && argv[i]; i++) {
    char *arg = argv[i];

    if('-' == *arg) {
      while('-' == *arg) arg++;
      switch( (int) *arg) {
        case 'q':
          log_queries = 1;
          break;
        case 'v':
          be_verbose = 1;
          break;
        case 'l':
          use_syslog = use_syslog ? 0 : 1;
          break;
        case 'h':
          usage(); exit(1);
        default :
          fprintf(stderr, "unknown argument!\n\n");
          usage(); exit(1);
      }
    } else {
      fprintf(stderr, "unknown argument!\n\n");
      usage(); exit(1);
    }
  }

  if(use_syslog) {
    openlog(appname, LOG_PID, LOG_AUTHPRIV);
  }

  signal(SIGINT,  quit);
  signal(SIGBUS,  quit);
  signal(SIGPIPE, quit);
  signal(SIGTRAP, quit);
  signal(SIGTERM, quit);
  signal(SIGQUIT, quit);
  signal(SIGKILL, quit);

  initConfig();
  readConfig(CONFIG_FILENAME);   /* use defaults if read fails */

  if(be_verbose) {
    if(use_syslog) {
      for(i=0; conf[i].key; i++) {
        syslog(LOG_DEBUG, "using %s => '%s'", conf[i].key,
                          conf[i].val ? conf[i].val : "");
      }
    } else {
      for(i=0; conf[i].key; i++) {
        fprintf(stderr, "%s[%d]: using %s => '%s'\n",
                        appname, getpid(), conf[i].key,
                        conf[i].val ? conf[i].val : "");
      }
    }
  }

  if(!(attrs = (char**)calloc((sizeof (char*)), 1))) {
    if(use_syslog) {
      syslog(LOG_ERR, "can not allocate memory");
    } else {
      fprintf(stderr, "%s[%d]: can not allocate memory\n",
                        appname, getpid());
    }
    exit(-1);
  }

  /*
   * apply config values
   */
  server = conf[0].val;
  if( conf[1].val) {
    port = atoi(conf[1].val);
  }
  suffix = conf[2].val;
  filter = conf[3].val;
  *attrs = conf[4].val;

  /*
   * connect to ldap server
   */
  if(!(ldap = ldapConnect(server, port, suffix))) {
    freeConfig();
    exit(-2);
  }

  if(be_verbose) {
    if(use_syslog) {
      syslog(LOG_DEBUG, "connection etablished - waiting for queries");
    } else {
      fprintf(stderr, "%s[%d]: connection etablished - waiting for queries\n",
                        appname, getpid());
    }
  }

  setbuf(stdout, NULL);
  while(fgets(buf, BUFSIZ-1, stdin)) {
    char *p = buf;
    if((p = strchr(p, '\n'))) *p = 0;
    if((p = strchr(p, '\r'))) *p = 0;

    if(!(user = strtok(buf, " "))) {
      squidReply(SQUID_ERR, user);
      continue;
    }
    if(!(pass = strtok(NULL, ""))) {
      squidReply(SQUID_ERR, user);
      continue;
    }

    if( ldapQueryPass(ldap, suffix, filter, attrs, user, pass)) {
      squidReply(SQUID_ERR, user);
    } else {
      squidReply(SQUID_OK, user);
    }

  }

  quit(0);
  return 0;
}

/* ---------------------------------------------------------------------- */
void  usage()
{
  fprintf(stderr, "usage: %s [-h] [-v] [-q] [-l]\n", appname);
  fprintf(stderr, "   -h   this help text\n");
  fprintf(stderr, "   -v   verbose mode - default is off\n");
  fprintf(stderr, "   -q   log queries - default is off\n");
  fprintf(stderr, "   -l   togle usage of syslog - default is on\n");
}

/* ---------------------------------------------------------------------- */
void quit(int status)
{
  if(status)
    squidReply(SQUID_ERR, "UNKNOWN");

  if(use_syslog) {
    if(status)
      syslog(LOG_DEBUG, "terminating on signal %d", status);
    closelog();
  } else {
    if(status)
      fprintf(stderr, "%s[%d]: terminating on signal %d\n",
                      appname, getpid(), status);
  }

  if(ldap)
    ldap_unbind(ldap);

  if(attrs)
    free(attrs);

  freeConfig();

  exit( status);
}

/* ---------------------------------------------------------------------- */
void  squidReply(char *reply, char *user)
{
  if(log_queries) {
    if(use_syslog) {
      syslog(LOG_NOTICE, "authentication request for '%s' - %s",
                         user ? user : "UNKNOWN", reply);
    } else {
      fprintf(stderr, "%s[%d]: authentication request for '%s' - %s",
                      appname, getpid(), user ? user : "UNKNOWN", reply);
    }
  }
  fputs(reply, stdout);
}

/* ---------------------------------------------------------------------- */
int  checkPass(char *user, char *spass, char *lpass) {
  char *pwd = lpass;
  int  rc = -1;

  if(!strncasecmp(pwd, ldap_crypt_prefix, crypt_prefix_size))
    pwd += crypt_prefix_size;

  if(strlen(pwd)>1 && '*' != *pwd && '!' != *pwd) {
    if( pwd != lpass) {
      // crypted pwd ?
      rc = strcmp(pwd, crypt(spass, pwd));
    } else {
      // plain text pwd ?!
      rc = strcmp(pwd, spass);
    }
  }

  return rc? -1: 0;
}


/* ---------------------------------------------------------------------- */
int   ldapQueryPass(LDAP *ldap, char *suffix, char *filter, 
                    char **attrs, char *user, char *pass)
{
  char           query[BUFSIZ] = {0};
  struct berval  **bvals =  0;
  LDAPMessage    *res, *e;
  BerElement     *ber;
  char           *a = 0;
  int            i, rc= 0;

  snprintf(query, sizeof(query), filter, user);
  if(-1 == ldap_search(ldap, suffix, LDAP_SCOPE_SUBTREE, query, attrs, 0)) {
    if(use_syslog) {
      syslog(LOG_ERR, "ldap search: %d", ldap->ld_errno);
    } else {
      fprintf(stderr, "%s[%d]: ldap search: %d\n",
                      appname, getpid(), ldap->ld_errno);
    }
    return -1;
  }

  while((rc = ldap_result(ldap, LDAP_RES_ANY, 0, NULL, &res))
        == LDAP_RES_SEARCH_ENTRY)
  {
    if(rc <= 0) {
      ldap_msgfree( res );
      return -1;
    }

    e = ldap_first_entry(ldap, res);
    if(e)    
    for(a = ldap_first_attribute(ldap, e, &ber); a != NULL;
        a = ldap_next_attribute(ldap, e, ber)) {
      if( !(bvals = ldap_get_values_len( ldap, e, a))) {
        if(be_verbose) {
          if(use_syslog) {
            syslog(LOG_ERR, "ldap entry= %s", a);
          } else {
            fprintf(stderr, "%s[%d]: ldap entry= %s\n",
                            appname, getpid(), a);
          }
        }
        ldap_msgfree( res);
        return -1;
      }

      if(be_verbose) {
        if(use_syslog) {
          for(i = 0; bvals[i]; i++)
            syslog(LOG_DEBUG, "ldap vals[%d]= '%s'", i, bvals[i]->bv_val);
        } else {
          for(i = 0; bvals[i]; i++)
            fprintf(stderr, "%s[%d]: ldap vals[%d]= '%s'\n",
                            appname, getpid(), i, bvals[i]->bv_val);
        }
      }
    }
  }

  /* we are using only first result! */
  rc = -1;
  if(bvals) {
    if(bvals[0]->bv_val) {
      rc = checkPass(user, pass, bvals[0]->bv_val);
    }
    ber_bvecfree( bvals);
  } else {
    if(be_verbose) {
      if(use_syslog) {
        syslog(LOG_DEBUG, "ldap result was empty (user not found)");
      } else {
        fprintf(stderr, "%s[%d]: ldap result was empty (user not found)\n",
                        appname, getpid());
      }
    }
  }
  ldap_msgfree( res);

  return rc;
}

/* ---------------------------------------------------------------------- */
LDAP* ldapConnect(char *server, int port, char *suffix)
{
  LDAP *ldap;

  if(!(ldap = ldap_init(server, port>0 && port<65535 ? port : LDAP_PORT))) {
    if(use_syslog) {
      syslog(LOG_NOTICE, "ldap_init failed");
    } else {
        fprintf(stderr, "%s[%d]: ldap_init failed\n",
                        appname, getpid());
    }
    return 0;
  }
  if(LDAP_SUCCESS != ldap_bind_s(ldap, conf[5].val ? conf[5].val : "",
                     conf[6].val ? conf[6].val : "", LDAP_AUTH_SIMPLE)) {
    if(use_syslog) {
      syslog(LOG_NOTICE, "ldap_bind failed");
    } else {
        fprintf(stderr, "%s[%d]: ldap_bind failed\n",
                        appname, getpid());
    }
    return 0;
  }
  return ldap;
}

/* ---------------------------------------------------------------------- */
void initConfig() {
  char buf[MAXHOSTNAMELEN+1];

  conf[0].val = strdup("localhost");
  conf[1].val = 0; // LDAP_PORT will be used

  if( !getdnsdomain(buf, sizeof(buf))) {
    int  dots   = 0;
    char *str   = 0;
    char *e, *b = buf;

    while(*b) if(*b++ == '.') dots++; // count dots in domain
    if(dots>0) {
      str = (char *)calloc(sizeof(char),strlen(buf)+strlen("dc=")*(dots+1)+1);
      if( str) {
        strcpy(str, "dc=");
        e = b = buf;
        while((e = strchr(b, '.'))) {
          strncat(str,b, e-b);
          strcat(str,",dc=");
          b=e+1;
        }
        strcat(str,b);
        conf[2].val = str;
      }
    }
  }
  if(!conf[2].val) {
    if(use_syslog) {
      syslog(LOG_NOTICE, "can not resolve local domainname");
    } else {
        fprintf(stderr, "%s[%d]: can not resolve local domainname\n",
                        appname, getpid());
    }
  }
  conf[3].val = strdup("(uid=%s)");
  conf[4].val = strdup("proxyauth");
}

/* ---------------------------------------------------------------------- */
void freeConfig()
{
  int i;
  for(i=0; conf[i].key; i++) {
    if(conf[i].val) free((void *)conf[i].val);
  }
}

/* ---------------------------------------------------------------------- */
int readConfig(char *filename)
{
  char  buf[BUFSIZ];
  char  *b=0, *e=0;
  int   i, len;
  FILE  *f;

  if(!(f=fopen(filename, "r"))) {
    if(be_verbose) {
      if(use_syslog) {
        syslog(LOG_WARNING, "can not read config file %s"
                            " - using defaults", filename);
      } else {
          fprintf(stderr, "%s[%d]: can not read config file %s"
                          " - using defaults\n",
                          appname, getpid(), filename);
      }
    }
    return 1;
  }

  while( !feof(f)) {
    if( !fgets(buf, sizeof(buf)-1, f))
      break;

    /* strip comments */
    e = buf;
    while((e = strchr(e, '#'))) {
      if(e>b) {
        if(*(e-1) == '\\') {
          e++; continue;
        }
      }
      *e = 0;
      break;
    }

    /* strip whitespaces on end */
    if((e = strrchr(buf, '\n'))) *e = 0;
    if((e = strrchr(buf, '\r'))) *e = 0;

    len = strlen(buf);
    e = buf+len-1;
    while(isspace(*e) && len>0) *e = 0, e--, len--;

    /* strip whitespaces on begin */
    b = buf;
    while(isspace(*b)) b++;

    /* search keyword */
    for(i=0; conf[i].key; i++) {

      len = strlen(conf[i].key);
      if(!strncasecmp(conf[i].key, b, len)) {
        if(be_verbose) {
          if(use_syslog) {
            syslog(LOG_DEBUG, "config - found key: '%s'", conf[i].key);
          } else {
            fprintf(stderr, "%s[%d]: config - found key: '%s'\n",
                            appname, getpid(), conf[i].key);
          }
        }

        /* search begin of the value */
        if( !(b = strchr(b+len, ':'))) break;
        b++;
        while(isspace(*b)) b++;

        if(be_verbose) {
          if(use_syslog) {
            syslog(LOG_DEBUG, "config - got value: '%s'", b);
          } else {
            fprintf(stderr, "%s[%d]: config - got value: '%s'\n",
                            appname, getpid(), b);
          }
        }

        /* copy value */
        len = strlen(b);
        if(len) {
          conf[i].val = (char *) calloc(sizeof(char), len+1);
          if(conf[i].val) {
            strncpy(conf[i].val, b, len); *(b+len) = 0;
          } else {
            if(use_syslog) {
              syslog(LOG_ERR, "can not allocate memory");
            } else {
              fprintf(stderr, "%s[%d]: can not allocate memory\n",
                              appname, getpid());
            }
          }
        }
      }
    }
  }
  fclose(f);

  return 0;
}

/* ---------------------------------------------------------------------- */
int getfqhostname(char *fqhost, size_t n)
{
  char           hname[MAXHOSTNAMELEN+1] = {0};
  struct hostent *hp;

  if( gethostname(hname, sizeof(hname)))
    return -1;

  if( !(hp = gethostbyname(hname))) {
    return -1;
  }
  strncpy(fqhost, hp->h_name, n-1);
  *(fqhost+n) = 0;

  return 0;
}

/* ---------------------------------------------------------------------- */
int getdnsdomain(char *domain, size_t n)
{
  char hname[MAXHOSTNAMELEN+1] = {0};
  char *p;

  if( getfqhostname(hname, sizeof(hname)))
    return -1;

  if( !(p = strchr(hname, '.')))
    return -1;

  strncpy(domain, p+1, n-1);
  *(domain+n) = 0;

  return 0;
}

/* -- EOF -- */

/*
 * $Log: squid_ldapauth.c,v $
 * Revision 1.3  2000/01/27 17:26:43  mt
 * added password authentication in ldap bind
 *
 * Revision 1.2  1999/11/09 13:14:27  mt
 * first public release
 *
 */

